@@ -1,4 +1,4 @@
-# Copyright 2004-2018 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2021 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -21,7 +21,8 @@
 
 # This contains various Displayables that handle events.
 
-from __future__ import print_function
+from __future__ import division, absolute_import, with_statement, print_function, unicode_literals
+from renpy.compat import *
 
 import renpy.display
 import renpy.audio
@@ -32,6 +33,8 @@ import pygame_sdl2 as pygame
 
 import math
 
+import re
+
 
 def compile_event(key, keydown):
     """
@@ -77,13 +80,7 @@ def compile_event(key, keydown):
     if part[0] == "joy" or part[0] == "pad":
         return "(False)"
 
-    # Otherwise, deal with it as a key.
-    if keydown:
-        rv = "(ev.type == %d" % pygame.KEYDOWN
-    else:
-        rv = "(ev.type == %d" % pygame.KEYUP
-
-    MODIFIERS = { "repeat", "alt", "meta", "shift", "noshift", "ctrl" }
+    MODIFIERS = { "keydown", "keyup", "repeat", "alt", "meta", "shift", "noshift", "ctrl" }
     modifiers = set()
 
     while part[0] in MODIFIERS:
@@ -91,6 +88,17 @@ def compile_event(key, keydown):
 
     key = "_".join(part)
 
+    if "keydown" in modifiers:
+        keydown = True
+    elif "keyup" in modifiers:
+        keydown = False
+
+    # Otherwise, deal with it as a key.
+    if keydown:
+        rv = "(ev.type == %d" % pygame.KEYDOWN
+    else:
+        rv = "(ev.type == %d" % pygame.KEYUP
+
     if "repeat" in modifiers:
         rv += " and (ev.repeat)"
     else:
@@ -216,6 +224,9 @@ def map_event(ev, keysym):
 
         return False
 
+    if isinstance(keysym, list):
+        keysym = tuple(keysym)
+
     check_code = event_cache.get(keysym, None)
     if check_code is None:
         check_code = eval("lambda ev : " + compile_event(keysym, True), globals())
@@ -224,17 +235,20 @@ def map_event(ev, keysym):
     return check_code(ev)
 
 
-def map_keyup(ev, name):
+def map_keyup(ev, keysym):
     """Returns true if the event matches the named keycode being released."""
 
     if ev.type == renpy.display.core.EVENTNAME:
-        if (name in ev.eventnames) and ev.up:
+        if (keysym in ev.eventnames) and ev.up:
             return True
 
-    check_code = keyup_cache.get(name, None)
+    if isinstance(keysym, list):
+        keysym = tuple(keysym)
+
+    check_code = keyup_cache.get(keysym, None)
     if check_code is None:
-        check_code = eval("lambda ev : " + compile_event(name, False), globals())
-        keyup_cache[name] = check_code
+        check_code = eval("lambda ev : " + compile_event(keysym, False), globals())
+        keyup_cache[keysym] = check_code
 
     return check_code(ev)
 
@@ -265,10 +279,10 @@ def skipping(ev):
 def inspector(ev):
     return map_event(ev, "inspector")
 
-
 ##############################################################################
 # Utility functions for dealing with actions.
 
+
 def predict_action(var):
     """
     Predicts some of the actions that may be caused by a variable.
@@ -295,7 +309,7 @@ def run(action, *args, **kwargs):
     arguments, a list of actions is run in order using this function, and
     None is ignored.
 
-    Returns the result of the first action to return a value.
+    Returns the result of the last action to return a value.
     """
 
     if action is None:
@@ -351,8 +365,26 @@ def run_periodic(var, st):
         return var.periodic(st)
 
 
+def get_tooltip(action):
+
+    if isinstance(action, (list, tuple)):
+        for i in action:
+            rv = get_tooltip(i)
+            if rv is not None:
+                return rv
+
+        return None
+
+    func = getattr(action, "get_tooltip", None)
+    if func is None:
+        return None
+
+    return func()
+
+
 def is_selected(action):
     """
+    :name: renpy.is_selected
     :doc: run
 
     Returns true if `action` indicates it is selected, or false otherwise.
@@ -360,7 +392,7 @@ def is_selected(action):
 
     if isinstance(action, (list, tuple)):
         for i in action:
-            if isinstance(i, renpy.store.SelectedIf):  # @UndefinedVariable
+            if isinstance(i, renpy.store.SelectedIf): # @UndefinedVariable
                 return i.get_selected()
         return any(is_selected(i) for i in action)
 
@@ -372,6 +404,7 @@ def is_selected(action):
 
 def is_sensitive(action):
     """
+    :name: renpy.is_sensitive
     :doc: run
 
     Returns true if `action` indicates it is sensitive, or False otherwise.
@@ -379,7 +412,7 @@ def is_sensitive(action):
 
     if isinstance(action, (list, tuple)):
         for i in action:
-            if isinstance(i, renpy.store.SensitiveIf):  # @UndefinedVariable
+            if isinstance(i, renpy.store.SensitiveIf): # @UndefinedVariable
                 return i.get_sensitive()
         return all(is_sensitive(i) for i in action)
 
@@ -430,7 +463,7 @@ class Keymap(renpy.display.layout.Null):
 
     def event(self, ev, x, y, st):
 
-        for name, action in self.keymap.iteritems():
+        for name, action in self.keymap.items():
 
             if map_event(ev, name):
 
@@ -444,7 +477,7 @@ class Keymap(renpy.display.layout.Null):
                 raise renpy.display.core.IgnoreEvent()
 
     def predict_one_action(self):
-        for i in self.keymap.itervalues():
+        for i in self.keymap.values():
             predict_action(i)
 
 
@@ -537,6 +570,9 @@ class SayBehavior(renpy.display.layout.Null):
         if not isinstance(dismiss, (list, tuple)):
             dismiss = [ dismiss ]
 
+        if not isinstance(dismiss_unfocused, (list, tuple)):
+            dismiss_unfocused = [ dismiss_unfocused ]
+
         if afm is not None:
             self.afm_length = len(afm)
         else:
@@ -544,6 +580,7 @@ class SayBehavior(renpy.display.layout.Null):
 
         # What keybindings lead to dismissal?
         self.dismiss = dismiss
+        self.dismiss_unfocused = dismiss_unfocused
 
         self.allow_dismiss = allow_dismiss
 
@@ -552,13 +589,19 @@ class SayBehavior(renpy.display.layout.Null):
 
     def set_text(self, text):
         self.text = text
-        self.afm_length = max(text.end - text.start, 1)
+
+        try:
+            afm_text = text.text[0][text.start:text.end]
+            afm_text = renpy.text.extras.filter_text_tags(afm_text, allow=[])
+            self.afm_length = max(len(afm_text), 1)
+        except:
+            self.afm_length = max(text.end - text.start, 1)
 
     def event(self, ev, x, y, st):
 
         if self.afm_length and renpy.game.preferences.afm_time and renpy.game.preferences.afm_enable:
 
-            afm_delay = ( 1.0 * ( renpy.config.afm_bonus + self.afm_length ) / renpy.config.afm_characters ) * renpy.game.preferences.afm_time
+            afm_delay = (1.0 * (renpy.config.afm_bonus + self.afm_length) / renpy.config.afm_characters) * renpy.game.preferences.afm_time
 
             if self.text is not None:
                 afm_delay += self.text.get_time()
@@ -590,7 +633,7 @@ class SayBehavior(renpy.display.layout.Null):
 
                 if not renpy.config.enable_rollback_side:
                     rollback_side = "disable"
-                if renpy.mobile:
+                if renpy.exports.mobile:
                     rollback_side = renpy.game.preferences.mobile_rollback_side
                 else:
                     rollback_side = renpy.game.preferences.desktop_rollback_side
@@ -637,7 +680,7 @@ class SayBehavior(renpy.display.layout.Null):
                 elif renpy.game.context().seen_current(True):
                     return True
                 else:
-                    renpy.config.skipping = False
+                    renpy.config.skipping = None
                     renpy.exports.restart_interaction()
 
             else:
@@ -645,10 +688,10 @@ class SayBehavior(renpy.display.layout.Null):
 
         return None
 
-
 ##############################################################################
 # Button
 
+
 KEY_EVENTS = (
     pygame.KEYDOWN,
     pygame.KEYUP,
@@ -672,6 +715,9 @@ class Button(renpy.display.layout.Window):
     keysym = None
     alternate_keysym = None
 
+    # This locks the displayable against further change.
+    locked = False
+
     def __init__(self, child=None, style='button', clicked=None,
                  hovered=None, unhovered=None, action=None, role=None,
                  time_policy=None, keymap={}, alternate=None,
@@ -691,7 +737,7 @@ class Button(renpy.display.layout.Window):
         self.unhovered = unhovered
         self.alternate = alternate
 
-        self.focusable = True  # (clicked is not None) or (action is not None)
+        self.focusable = True # (clicked is not None) or (action is not None)
         self.role_parameter = role
 
         self.keymap = keymap
@@ -704,8 +750,25 @@ class Button(renpy.display.layout.Window):
         self._duplicatable = False
 
     def _duplicate(self, args):
+        if args and args.args:
+            args.extraneous()
+
         return self
 
+    def _get_tooltip(self):
+        if self._tooltip is not None:
+            return self._tooltip
+
+        return get_tooltip(self.action)
+
+    def _in_current_store(self):
+        rv = self._copy()
+        rv.style = self.style.copy()
+        rv.set_style_prefix(self.style.prefix, True)
+        rv.focusable = False
+        rv.locked = True
+        return rv
+
     def predict_one_action(self):
         predict_action(self.clicked)
         predict_action(self.hovered)
@@ -713,7 +776,7 @@ class Button(renpy.display.layout.Window):
         predict_action(self.alternate)
 
         if self.keymap:
-            for v in self.keymap.itervalues():
+            for v in self.keymap.values():
                 predict_action(v)
 
     def render(self, width, height, st, at):
@@ -801,40 +864,46 @@ class Button(renpy.display.layout.Window):
 
     def per_interact(self):
 
-        if self.action is not None:
-            if self.is_selected():
-                role = 'selected_'
+        if not self.locked:
+
+            if self.action is not None:
+                if self.is_selected():
+                    role = 'selected_'
+                else:
+                    role = ''
+
+                if self.is_sensitive():
+                    clicked = self.action
+                else:
+                    clicked = None
+                    role = ''
+
             else:
                 role = ''
+                clicked = self.clicked
 
-            if self.is_sensitive():
-                clicked = self.action
+            if self.role_parameter is not None:
+                role = self.role_parameter
+
+            if (role != self.role) or (clicked is not self.clicked):
+                renpy.display.render.invalidate(self)
+                self.role = role
+                self.clicked = clicked
+
+            if self.clicked is not None:
+                self.set_style_prefix(self.role + "idle_", True)
+                self.focusable = True
             else:
-                clicked = None
-
-        else:
-            role = ''
-            clicked = self.clicked
-
-        if self.role_parameter is not None:
-            role = self.role_parameter
-
-        if (role != self.role) or (clicked is not self.clicked):
-            renpy.display.render.invalidate(self)
-            self.role = role
-            self.clicked = clicked
-
-        if self.clicked is not None:
-            self.set_style_prefix(self.role + "idle_", True)
-            self.focusable = True
-        else:
-            self.set_style_prefix(self.role + "insensitive_", True)
-            self.focusable = False
+                self.set_style_prefix(self.role + "insensitive_", True)
+                self.focusable = False
 
         super(Button, self).per_interact()
 
     def event(self, ev, x, y, st):
 
+        if self.locked:
+            return None
+
         def handle_click(action):
             renpy.exports.play(self.style.activate_sound)
 
@@ -871,7 +940,7 @@ class Button(renpy.display.layout.Window):
             return None
 
         # Check the keymap.
-        for name, action in self.keymap.iteritems():
+        for name, action in self.keymap.items():
             if map_event(ev, name):
                 return run(action)
 
@@ -898,7 +967,7 @@ class Button(renpy.display.layout.Window):
         if (self.clicked is not None) and map_event(ev, "button_ignore"):
             raise renpy.display.core.IgnoreEvent()
 
-        if (self.clicked is not None) and map_event(ev, "button_alternate_ignore"):
+        if (self.alternate is not None) and map_event(ev, "button_alternate_ignore"):
             raise renpy.display.core.IgnoreEvent()
 
         # If clicked,
@@ -932,8 +1001,10 @@ class Button(renpy.display.layout.Window):
 def TextButton(text, style='button', text_style='button_text',
                clicked=None, **properties):
 
-    text = renpy.text.text.Text(text, style=text_style)  # @UndefinedVariable
-    return Button(text, style=style, clicked=clicked, **properties)
+    text_properties, button_properties = renpy.easy.split_properties(properties, "text_", "")
+
+    text = renpy.text.text.Text(text, style=text_style, **text_properties) # @UndefinedVariable
+    return Button(text, style=style, clicked=clicked, **button_properties)
 
 
 class ImageButton(Button):
@@ -983,7 +1054,7 @@ class ImageButton(Button):
                                           **properties)
 
     def visit(self):
-        return self.state_children.values()
+        return list(self.state_children.values())
 
     def get_child(self):
         return self.style.child or self.state_children[self.style.prefix]
@@ -1055,7 +1126,7 @@ def input_post_per_interact():
             i.caret_pos = len(content)
 
 
-class Input(renpy.text.text.Text):  # @UndefinedVariable
+class Input(renpy.text.text.Text): # @UndefinedVariable
     """
     This is a Displayable that takes text as input.
     """
@@ -1069,6 +1140,7 @@ class Input(renpy.text.text.Text):  # @UndefinedVariable
     default = u""
     edit_text = u""
     value = None
+    shown = False
 
     def __init__(self,
                  default="",
@@ -1084,6 +1156,7 @@ class Input(renpy.text.text.Text):  # @UndefinedVariable
                  editable=True,
                  pixel_width=None,
                  value=None,
+                 copypaste=False,
                  **properties):
 
         super(Input, self).__init__("", style=style, replaces=replaces, substitute=False, **properties)
@@ -1093,7 +1166,7 @@ class Input(renpy.text.text.Text):  # @UndefinedVariable
             changed = value.set_text
             default = value.get_text()
 
-        self.default = unicode(default)
+        self.default = str(default)
         self.content = self.default
 
         self.length = length
@@ -1102,6 +1175,7 @@ class Input(renpy.text.text.Text):  # @UndefinedVariable
         self.exclude = exclude
         self.prefix = prefix
         self.suffix = suffix
+        self.copypaste = copypaste
 
         self.changed = changed
 
@@ -1127,15 +1201,10 @@ class Input(renpy.text.text.Text):  # @UndefinedVariable
             self.content = replaces.content
             self.editable = replaces.editable
             self.caret_pos = replaces.caret_pos
+            self.shown = replaces.shown
 
         self.update_text(self.content, self.editable)
 
-    def _show(self):
-        if self.default != self.content:
-            self.content = self.default
-            self.caret_pos = len(self.content)
-            self.update_text(self.content, self.editable)
-
     def update_text(self, new_content, editable, check_size=False):
 
         edit = renpy.display.interface.text_editing
@@ -1228,6 +1297,18 @@ class Input(renpy.text.text.Text):  # @UndefinedVariable
             if self.value.default and (default_input_value is None):
                 default_input_value = self.value
 
+        if not self.shown:
+
+            if self.value is not None:
+                default = self.value.get_text()
+                self.default = str(default)
+
+            self.content = self.default
+            self.caret_pos = len(self.content)
+            self.update_text(self.content, self.editable)
+
+            self.shown = True
+
     def event(self, ev, x, y, st):
 
         self.old_caret_pos = self.caret_pos
@@ -1245,7 +1326,7 @@ class Input(renpy.text.text.Text):  # @UndefinedVariable
         if map_event(ev, "input_backspace"):
 
             if self.content and self.caret_pos > 0:
-                content = self.content[0:self.caret_pos-1] + self.content[self.caret_pos:l]
+                content = self.content[0:self.caret_pos - 1] + self.content[self.caret_pos:l]
                 self.caret_pos -= 1
                 self.update_text(content, self.editable)
 
@@ -1283,7 +1364,7 @@ class Input(renpy.text.text.Text):  # @UndefinedVariable
 
         elif map_event(ev, "input_delete"):
             if self.caret_pos < l:
-                content = self.content[0:self.caret_pos] + self.content[self.caret_pos+1:l]
+                content = self.content[0:self.caret_pos] + self.content[self.caret_pos + 1:l]
                 self.update_text(content, self.editable)
 
             renpy.display.render.redraw(self, 0)
@@ -1301,6 +1382,19 @@ class Input(renpy.text.text.Text):  # @UndefinedVariable
             renpy.display.render.redraw(self, 0)
             raise renpy.display.core.IgnoreEvent()
 
+        elif self.copypaste and map_event(ev, "input_copy"):
+            text = self.content.encode("utf-8")
+            pygame.scrap.put(pygame.scrap.SCRAP_TEXT, text)
+            raise renpy.display.core.IgnoreEvent()
+
+        elif self.copypaste and map_event(ev, "input_paste"):
+            text = pygame.scrap.get(pygame.scrap.SCRAP_TEXT)
+            text = text.decode("utf-8")
+            raw_text = ""
+            for c in text:
+                if ord(c) >= 32:
+                    raw_text += c
+
         elif ev.type == pygame.TEXTEDITING:
             self.update_text(self.content, self.editable, check_size=True)
 
@@ -1315,7 +1409,10 @@ class Input(renpy.text.text.Text):  # @UndefinedVariable
             if ev.unicode and ord(ev.unicode[0]) >= 32:
                 raw_text = ev.unicode
             elif renpy.display.interface.text_event_in_queue():
-                raw_text = ''
+                raise renpy.display.core.IgnoreEvent()
+            elif (32 <= ev.key < 127) and not (ev.mod & (pygame.KMOD_ALT | pygame.KMOD_META)):
+                # Ignore printable keycodes without unicode.
+                raise renpy.display.core.IgnoreEvent()
 
         if raw_text is not None:
 
@@ -1323,10 +1420,33 @@ class Input(renpy.text.text.Text):  # @UndefinedVariable
 
             for c in raw_text:
 
-                if self.allow and c not in self.allow:
-                    continue
-                if self.exclude and c in self.exclude:
-                    continue
+                # Allow is given
+                if self.allow:
+
+                    # Allow is regex
+                    if isinstance(self.allow, re.Pattern):
+
+                        # Character doesn't match
+                        if self.allow.search(c) is None:
+                            continue
+
+                    # Allow is string
+                    elif c not in self.allow:
+                        continue
+
+                # Exclude is given
+                if self.exclude:
+
+                    # Exclude is regex
+                    if isinstance(self.exclude, re.Pattern):
+
+                        # Character matches
+                        if self.exclude.search(c) is not None:
+                            continue
+
+                    # Exclude is string
+                    elif c in self.exclude:
+                        continue
 
                 text += c
 
@@ -1373,7 +1493,9 @@ class Adjustment(renpy.object.Object):
 
     """
 
-    def __init__(self, range=1, value=0, step=None, page=None, changed=None, adjustable=None, ranged=None):  # @ReservedAssignment
+    force_step = False
+
+    def __init__(self, range=1, value=0, step=None, page=None, changed=None, adjustable=None, ranged=None, force_step=False): # @ReservedAssignment
         """
         The following parameters correspond to fields or properties on
         the adjustment object:
@@ -1401,12 +1523,12 @@ class Adjustment(renpy.object.Object):
         The following parameters control the behavior of the adjustment.
 
         `adjustable`
-             If True, this adjustment can be changed by a bar. If False,
-             it can't.
+            If True, this adjustment can be changed by a bar. If False,
+            it can't.
 
-             It defaults to being adjustable if a `changed` function
-             is given or if the adjustment is associated with a viewport,
-             and not adjustable otherwise.
+            It defaults to being adjustable if a `changed` function
+            is given or if the adjustment is associated with a viewport,
+            and not adjustable otherwise.
 
         `changed`
             This function is called with the new value when the value of
@@ -1416,6 +1538,16 @@ class Adjustment(renpy.object.Object):
             This function is called with the adjustment object when
             the range of the adjustment is set by a viewport.
 
+        `force_step`
+            If True and this adjustment changes by dragging associated
+            viewport or a bar, value will be changed only if the drag
+            reached next step.
+            If "release" and this adjustment changes by dragging associated
+            viewport or a bar, after the release, value will be
+            rounded to the nearest step.
+            If False, this adjustment will changes by dragging, ignoring
+            the step value.
+
         .. method:: change(value)
 
             Changes the value of the adjustment to `value`, updating
@@ -1428,16 +1560,34 @@ class Adjustment(renpy.object.Object):
             if changed:
                 adjustable = True
 
-        self._value = value
         self._range = range
+        self._value = type(range)(value)
         self._page = page
         self._step = step
         self.changed = changed
         self.adjustable = adjustable
         self.ranged = ranged
+        self.force_step = force_step
+
+    def round_value(self, value, release):
+        # Prevent deadlock border points
+        if value <= 0:
+            return type(self._value)(0)
+        elif value >= self._range:
+            return self._range
+
+        if self.force_step is False:
+            return value
+
+        if (not release) and self.force_step == "release":
+            return value
+
+        return type(self._value)(self.step * round(float(value) / self.step))
 
     def get_value(self):
-        if self._value > self._range:
+        if self._value <= 0:
+            return type(self._value)(0)
+        if self._value >= self._range:
             return self._range
 
         return self._value
@@ -1455,7 +1605,7 @@ class Adjustment(renpy.object.Object):
         if self.ranged:
             self.ranged(self)
 
-    range = property(get_range, set_range)  # @ReservedAssignment
+    range = property(get_range, set_range) # @ReservedAssignment
 
     def get_page(self):
         if self._page is not None:
@@ -1512,7 +1662,7 @@ class Adjustment(renpy.object.Object):
         """
 
         for d in adj_registered.setdefault(self, [ ]):
-            renpy.display.render.redraw(d, 0)
+            renpy.display.render.invalidate(d)
 
 
 class Bar(renpy.display.core.Displayable):
@@ -1526,17 +1676,17 @@ class Bar(renpy.display.core.Displayable):
     def after_upgrade(self, version):
 
         if version < 1:
-            self.adjustment = Adjustment(self.range, self.value, changed=self.changed)  # E1101
+            self.adjustment = Adjustment(self.range, self.value, changed=self.changed) # E1101
             self.adjustment.register(self)
-            del self.range  # E1101
-            del self.value  # E1101
-            del self.changed  # E1101
+            del self.range # E1101
+            del self.value # E1101
+            del self.changed # E1101
 
         if version < 2:
             self.value = None
 
     def __init__(self,
-                 range=None,  # @ReservedAssignment
+                 range=None, # @ReservedAssignment
                  value=None,
                  width=None,
                  height=None,
@@ -1564,6 +1714,10 @@ class Bar(renpy.display.core.Displayable):
                 adjustment = value.get_adjustment()
                 renpy.game.interface.timeout(0)
 
+                tooltip = value.get_tooltip()
+                if tooltip is not None:
+                    properties.setdefault("tooltip", tooltip)
+
             else:
                 adjustment = Adjustment(range, value, step=step, page=page, changed=changed)
 
@@ -1611,9 +1765,10 @@ class Bar(renpy.display.core.Displayable):
         self.focusable = self.adjustment.adjustable
         self.adjustment.register(self)
 
-    def predict_one(self):
-        pd = renpy.display.predict.displayable
-        self.style._predict_window(pd)
+    def visit(self):
+        rv = [ ]
+        self.style._visit_bar(rv.append)
+        return rv
 
     def render(self, width, height, st, at):
 
@@ -1634,7 +1789,7 @@ class Bar(renpy.display.core.Displayable):
         # Store the width and height for the event function to use.
         self.width = width
         self.height = height
-        range = self.adjustment.range  # @ReservedAssignment
+        range = self.adjustment.range # @ReservedAssignment
         value = self.adjustment.value
         page = self.adjustment.page
 
@@ -1662,7 +1817,7 @@ class Bar(renpy.display.core.Displayable):
 
         active = dimension - fore_gutter - aft_gutter
         if range:
-            thumb_dim = active * page / (range + page)
+            thumb_dim = active * page // (range + page)
         else:
             thumb_dim = active
 
@@ -1684,7 +1839,7 @@ class Bar(renpy.display.core.Displayable):
         active -= thumb_dim
 
         if range:
-            fore_size = active * value / range
+            fore_size = active * value // range
         else:
             fore_size = active
 
@@ -1704,7 +1859,7 @@ class Bar(renpy.display.core.Displayable):
                 aftsurf = render(self.style.aft_bar, width, aft_size, st, at)
                 rv.blit(thumb_shadow, (0, fore_size - thumb_offset))
                 rv.blit(foresurf, (0, 0), main=False)
-                rv.blit(aftsurf, (0, height-aft_size), main=False)
+                rv.blit(aftsurf, (0, height - aft_size), main=False)
                 rv.blit(thumb, (0, fore_size - thumb_offset))
 
             else:
@@ -1722,7 +1877,7 @@ class Bar(renpy.display.core.Displayable):
                 aftsurf = render(self.style.aft_bar, aft_size, height, st, at)
                 rv.blit(thumb_shadow, (fore_size - thumb_offset, 0))
                 rv.blit(foresurf, (0, 0), main=False)
-                rv.blit(aftsurf, (width-aft_size, 0), main=False)
+                rv.blit(aftsurf, (width - aft_size, 0), main=False)
                 rv.blit(thumb, (fore_size - thumb_offset, 0))
 
             else:
@@ -1731,7 +1886,7 @@ class Bar(renpy.display.core.Displayable):
 
                 rv.blit(thumb_shadow, (fore_size - thumb_offset, 0))
                 rv.blit(foresurf.subsurface((0, 0, fore_size, height)), (0, 0), main=False)
-                rv.blit(aftsurf.subsurface((width - aft_size, 0, aft_size, height)), (width-aft_size, 0), main=False)
+                rv.blit(aftsurf.subsurface((width - aft_size, 0, aft_size, height)), (width - aft_size, 0), main=False)
                 rv.blit(thumb, (fore_size - thumb_offset, 0))
 
         if self.focusable:
@@ -1765,7 +1920,7 @@ class Bar(renpy.display.core.Displayable):
         if self.hidden:
             return None
 
-        range = self.adjustment.range  # @ReservedAssignment
+        range = self.adjustment.range # @ReservedAssignment
         old_value = self.adjustment.value
         value = old_value
 
@@ -1847,9 +2002,18 @@ class Bar(renpy.display.core.Displayable):
             renpy.display.tts.speak(renpy.minstore.__("deactivate"))
             self.set_style_prefix("hover_", True)
             renpy.display.focus.set_grab(None)
-            ignore_event = True
+
+            # Invoke rounding adjustment on bar release
+            value = self.adjustment.round_value(value, release=True)
+            if value != old_value:
+                rv = self.adjustment.change(value)
+                if rv is not None:
+                    return rv
+
+            raise renpy.display.core.IgnoreEvent()
 
         if value != old_value:
+            value = self.adjustment.round_value(value, release=False)
             rv = self.adjustment.change(value)
             if rv is not None:
                 return rv
@@ -1912,13 +2076,11 @@ class Conditional(renpy.display.layout.Container):
             return self.child.event(ev, x, y, st)
 
 
-class TimerState(renpy.python.RevertableObject):
+class TimerState(renpy.python.AlwaysRollback):
     """
     Stores the state of the timer, which may need to be rolled back.
     """
 
-    # Prevents us from having to worry about our initialization being
-    # rolled back.
     started = False
     next_event = None
 
@@ -1938,9 +2100,6 @@ class Timer(renpy.display.layout.Null):
     def __init__(self, delay, action=None, repeat=False, args=(), kwargs={}, replaces=None, **properties):
         super(Timer, self).__init__(**properties)
 
-        if action is None:
-            raise Exception("A timer must have an action supplied.")
-
         if delay <= 0:
             raise Exception("A timer's delay must be > 0.")
 
@@ -2028,14 +2187,18 @@ class MouseArea(renpy.display.core.Displayable):
         if renpy.display.focus.get_grab():
             return
 
-        if self.style.focus_mask is not None:
-            crend = renpy.display.render.render(self.style.focus_mask, self.width, self.height, st, self.at_st_offset + st)
-            is_hovered = crend.is_pixel_opaque(x, y)
-        elif 0 <= x < self.width and 0 <= y < self.height:
-            is_hovered = True
-        else:
+        if renpy.display.focus.pending_focus_type == 'keyboard':
             is_hovered = False
 
+        else:
+            if self.style.focus_mask is not None:
+                crend = renpy.display.render.render(self.style.focus_mask, self.width, self.height, st, self.at_st_offset + st)
+                is_hovered = crend.is_pixel_opaque(x, y)
+            elif 0 <= x < self.width and 0 <= y < self.height:
+                is_hovered = True
+            else:
+                is_hovered = False
+
         if is_hovered and not self.is_hovered:
             self.is_hovered = True
 
@@ -2068,14 +2231,20 @@ class OnEvent(renpy.display.core.Displayable):
         self.event_name = event
         self.action = action
 
+    def is_event(self, event):
+        if isinstance(self.event_name, basestring):
+            return self.event_name == event
+        else:
+            return event in self.event_name
+
     def _handles_event(self, event):
-        if self.event_name == event:
+        if self.is_event(event):
             return True
         else:
             return False
 
     def set_transform_event(self, event):
-        if event == self.event_name:
+        if self.is_event(event):
             run(self.action)
 
     def render(self, width, height, st, at):
